@@ -23,17 +23,21 @@
 
 #ifdef PY3K
 #define INTERN PyUnicode_InternFromString
+#define INTERN_INPLACE PyUnicode_InternInPlace
+#define NATIVE_CHECK_EXACT PyUnicode_CheckExact
 #define NATIVE_FROM_STRING_AND_SIZE PyUnicode_FromStringAndSize
 
 #define Py_TPFLAGS_HAVE_RICHCOMPARE 0
 
 #define INT_FROM_LONG(x) PyLong_FromLong(x)
 #define INT_CHECK(x) PyLong_Check(x)
-#define INT_AS_LONG(x) PyLong_AS_LONG(x)
+#define INT_AS_LONG(x) PyLong_AsLong(x)
 #define CAPI_CAPSULE_NAME "persistent.cPersistence.CAPI"
 
 #else
 #define INTERN PyString_InternFromString
+#define INTERN_INPLACE PyString_InternInPlace
+#define NATIVE_CHECK_EXACT PyString_CheckExact
 #define NATIVE_FROM_STRING_AND_SIZE PyString_FromStringAndSize
 
 #define INT_FROM_LONG(x) PyInt_FromLong(x)
