@@ -11,19 +11,31 @@
 # FOR A PARTICULAR PURPOSE.
 #
 ##############################################################################
-import os
+
+import re
 import unittest
 
-import platform
-import sys
-py_impl = getattr(platform, 'python_implementation', lambda: None)
-_is_pypy3 = py_impl() == 'PyPy' and sys.version_info[0] > 2
-_is_jython = py_impl() == 'Jython'
+from persistent._compat import copy_reg
+from persistent._compat import PYPY
+from persistent._compat import PYTHON3 as PY3
+from persistent.tests.utils import skipIfNoCExtension
+
+_is_pypy3 = PYPY and PY3
+
+# pylint:disable=R0904,W0212,E1101
+# pylint:disable=attribute-defined-outside-init,too-many-lines
+# pylint:disable=blacklisted-name,useless-object-inheritance
+# Hundreds of unused jar and OID vars make this useless
+# pylint:disable=unused-variable
 
-#pylint: disable=R0904,W0212,E1101
 
 class _Persistent_Base(object):
 
+    # py2/3 compat
+    assertRaisesRegex = getattr(unittest.TestCase,
+                                'assertRaisesRegex',
+                                unittest.TestCase.assertRaisesRegexp)
+
     def _getTargetClass(self):
         # concrete testcase classes must override
         raise NotImplementedError()
@@ -32,6 +44,9 @@ class _Persistent_Base(object):
         # concrete testcase classes must override
         raise NotImplementedError()
 
+    def _makeRealCache(self, jar):
+        return self._makeCache(jar)
+
     def _makeOne(self, *args, **kw):
         return self._getTargetClass()(*args, **kw)
 
@@ -73,24 +88,23 @@ class _Persistent_Base(object):
         class _BrokenJar(object):
             def __init__(self):
                 self.called = 0
-            def register(self,ob):
+            def register(self, ob):
                 self.called += 1
-                raise NotImplementedError
-            def setstate(self,ob):
-                raise NotImplementedError
+                raise NotImplementedError()
+            def setstate(self, ob):
+                raise NotImplementedError()
 
         jar = _BrokenJar()
         jar._cache = self._makeCache(jar)
         return jar
 
-    def _makeOneWithJar(self, klass=None):
-        from persistent.timestamp import _makeOctets
-        OID = _makeOctets('\x01' * 8)
+    def _makeOneWithJar(self, klass=None, broken_jar=False):
+        OID = b'\x01' * 8
         if klass is not None:
             inst = klass()
         else:
             inst = self._makeOne()
-        jar = self._makeJar()
+        jar = self._makeJar() if not broken_jar else self._makeBrokenJar()
         jar._cache.new_ghost(OID, inst) # assigns _p_jar, _p_oid
         return inst, jar, OID
 
@@ -104,6 +118,14 @@ class _Persistent_Base(object):
         from persistent.interfaces import IPersistent
         verifyObject(IPersistent, self._makeOne())
 
+    def test_instance_cannot_be_weakly_referenced(self):
+        if PYPY: # pragma: no cover
+            self.skipTest('On PyPy, everything can be weakly referenced')
+        import weakref
+        inst = self._makeOne()
+        with self.assertRaises(TypeError):
+            weakref.ref(inst)
+
     def test_ctor(self):
         from persistent.persistence import _INITIAL_SERIAL
         inst = self._makeOne()
@@ -156,12 +178,10 @@ class _Persistent_Base(object):
     def test_assign_p_jar_w_new_jar(self):
         inst, jar, OID = self._makeOneWithJar()
         new_jar = self._makeJar()
-        try:
+
+        with self.assertRaisesRegex(ValueError,
+                                    "can not change _p_jar of cached object"):
             inst._p_jar = new_jar
-        except ValueError as e:
-            self.assertEqual(str(e), "can not change _p_jar of cached object")
-        else:
-            self.fail("Should raise ValueError")
 
     def test_assign_p_jar_w_valid_jar(self):
         jar = self._makeJar()
@@ -183,43 +203,35 @@ class _Persistent_Base(object):
     def test_assign_p_oid_w_invalid_oid(self):
         inst, jar, OID = self._makeOneWithJar()
 
-        try:
+        with self.assertRaisesRegex(ValueError,
+                                    'can not change _p_oid of cached object'):
             inst._p_oid = object()
-        except ValueError as e:
-            self.assertEqual(str(e), 'can not change _p_oid of cached object')
-        else:
-            self.fail("Should raise value error")
-
 
     def test_assign_p_oid_w_valid_oid(self):
-        from persistent.timestamp import _makeOctets
-        OID = _makeOctets('\x01' * 8)
+        OID = b'\x01' * 8
         inst = self._makeOne()
         inst._p_oid = OID
         self.assertEqual(inst._p_oid, OID)
         inst._p_oid = OID  # reassign only same OID
 
     def test_assign_p_oid_w_new_oid_wo_jar(self):
-        from persistent.timestamp import _makeOctets
-        OID1 = _makeOctets('\x01' * 8)
-        OID2 = _makeOctets('\x02' * 8)
+        OID1 = b'\x01' * 8
+        OID2 = b'\x02' * 8
         inst = self._makeOne()
         inst._p_oid = OID1
         inst._p_oid = OID2
         self.assertEqual(inst._p_oid, OID2)
 
     def test_assign_p_oid_w_None_wo_jar(self):
-        from persistent.timestamp import _makeOctets
-        OID1 = _makeOctets('\x01' * 8)
+        OID1 = b'\x01' * 8
         inst = self._makeOne()
         inst._p_oid = OID1
         inst._p_oid = None
         self.assertEqual(inst._p_oid, None)
 
     def test_assign_p_oid_w_new_oid_w_jar(self):
-        from persistent.timestamp import _makeOctets
         inst, jar, OID = self._makeOneWithJar()
-        new_OID = _makeOctets('\x02' * 8)
+        new_OID = b'\x02' * 8
         def _test():
             inst._p_oid = new_OID
         self.assertRaises(ValueError, _test)
@@ -233,8 +245,7 @@ class _Persistent_Base(object):
         self.assertEqual(inst._p_oid, 42)
 
     def test_delete_p_oid_wo_jar(self):
-        from persistent.timestamp import _makeOctets
-        OID = _makeOctets('\x01' * 8)
+        OID = b'\x01' * 8
         inst = self._makeOne()
         inst._p_oid = OID
         del inst._p_oid
@@ -242,9 +253,18 @@ class _Persistent_Base(object):
 
     def test_delete_p_oid_w_jar(self):
         inst, jar, OID = self._makeOneWithJar()
-        def _test():
+        with self.assertRaises(ValueError):
+            del inst._p_oid
+
+    def test_delete_p_oid_of_subclass_calling_p_delattr(self):
+        class P(self._getTargetClass()):
+            def __delattr__(self, name):
+                super(P, self)._p_delattr(name)
+                raise AssertionError("Should not get here")
+
+        inst, _jar, _oid = self._makeOneWithJar(klass=P)
+        with self.assertRaises(ValueError):
             del inst._p_oid
-        self.assertRaises(ValueError, _test)
 
     def test_del_oid_like_ZODB_abort(self):
         # When a ZODB connection aborts, it removes registered objects from
@@ -270,30 +290,28 @@ class _Persistent_Base(object):
     def test_assign_p_serial_too_short(self):
         inst = self._makeOne()
         def _test():
-            inst._p_serial = '\x01\x02\x03'
+            inst._p_serial = b'\x01\x02\x03'
         self.assertRaises(ValueError, _test)
 
     def test_assign_p_serial_too_long(self):
         inst = self._makeOne()
         def _test():
-            inst._p_serial = '\x01\x02\x03' * 3
+            inst._p_serial = b'\x01\x02\x03' * 3
         self.assertRaises(ValueError, _test)
 
     def test_assign_p_serial_w_valid_serial(self):
-        from persistent.timestamp import _makeOctets
-        SERIAL = _makeOctets('\x01' * 8)
+        SERIAL = b'\x01' * 8
         inst = self._makeOne()
         inst._p_serial = SERIAL
         self.assertEqual(inst._p_serial, SERIAL)
 
     def test_delete_p_serial(self):
-        from persistent.timestamp import _makeOctets
         from persistent.persistence import _INITIAL_SERIAL
-        SERIAL = _makeOctets('\x01' * 8)
+        SERIAL = b'\x01' * 8
         inst = self._makeOne()
         inst._p_serial = SERIAL
         self.assertEqual(inst._p_serial, SERIAL)
-        del(inst._p_serial)
+        del inst._p_serial
         self.assertEqual(inst._p_serial, _INITIAL_SERIAL)
 
     def test_query_p_changed_unsaved(self):
@@ -617,15 +635,17 @@ class _Persistent_Base(object):
 
     def test_assign_p_estimated_size_wrong_type(self):
         inst = self._makeOne()
-        self.assertRaises(TypeError,
-                          lambda : setattr(inst, '_p_estimated_size', None))
+
+        with self.assertRaises(TypeError):
+            inst._p_estimated_size = None
+
         try:
-            long
+            constructor = long
         except NameError:
-            pass
-        else:
-            self.assertRaises(TypeError,
-                          lambda : setattr(inst, '_p_estimated_size', long(1)))
+            constructor = str
+
+        with self.assertRaises(TypeError):
+            inst._p_estimated_size = constructor(1)
 
     def test_assign_p_estimated_size_negative(self):
         inst = self._makeOne()
@@ -717,7 +737,7 @@ class _Persistent_Base(object):
             def __getattribute__(self, name):
                 if name == 'magic':
                     return 42
-                return super(Base,self).__getattribute__(name)
+                return super(Base, self).__getattribute__(name) # pragma: no cover
 
         self.assertEqual(getattr(Base(), 'magic'), 42)
 
@@ -727,8 +747,7 @@ class _Persistent_Base(object):
         self.assertRaises(AttributeError, getattr, Derived(), 'magic')
 
     def test___setattr___p__names(self):
-        from persistent.timestamp import _makeOctets
-        SERIAL = _makeOctets('\x01' * 8)
+        SERIAL = b'\x01' * 8
         inst, jar, OID = self._makeOneWithJar()
         inst._p_activate()
         NAMES = [('_p_jar', jar),
@@ -915,7 +934,7 @@ class _Persistent_Base(object):
         from persistent.persistence import _INITIAL_SERIAL
         inst = self._makeOne()
         self.assertRaises((ValueError, TypeError),
-                           inst.__setstate__, {'bogus': 1})
+                          inst.__setstate__, {'bogus': 1})
         self.assertEqual(inst._p_jar, None)
         self.assertEqual(inst._p_oid, None)
         self.assertEqual(inst._p_serial, _INITIAL_SERIAL)
@@ -967,7 +986,7 @@ class _Persistent_Base(object):
         self.assertEqual(inst.baz, 'bam')
         self.assertEqual(inst.qux, 'spam')
 
-    if not _is_pypy3 and not _is_jython:
+    if not _is_pypy3:
         def test___setstate___interns_dict_keys(self):
             class Derived(self._getTargetClass()):
                 pass
@@ -982,6 +1001,23 @@ class _Persistent_Base(object):
             key2 = list(inst2.__dict__.keys())[0]
             self.assertTrue(key1 is key2)
 
+            from persistent._compat import IterableUserDict
+            inst1 = Derived()
+            inst2 = Derived()
+            key1 = 'key'
+            key2 = 'ke'; key2 += 'y'  # construct in a way that won't intern the literal
+            self.assertFalse(key1 is key2)
+            state1 = IterableUserDict({key1: 1})
+            state2 = IterableUserDict({key2: 2})
+            k1 = list(state1.keys())[0]
+            k2 = list(state2.keys())[0]
+            self.assertFalse(k1 is k2)  # verify
+            inst1.__setstate__(state1)
+            inst2.__setstate__(state2)
+            key1 = list(inst1.__dict__.keys())[0]
+            key2 = list(inst2.__dict__.keys())[0]
+            self.assertTrue(key1 is key2)
+
     def test___setstate___doesnt_fail_on_non_string_keys(self):
         class Derived(self._getTargetClass()):
             pass
@@ -995,8 +1031,18 @@ class _Persistent_Base(object):
         inst1.__setstate__({mystr: 2})
         self.assertTrue(mystr in inst1.__dict__)
 
+    def test___setstate___doesnt_fail_on_non_dict(self):
+        class Derived(self._getTargetClass()):
+            pass
+        inst1 = Derived()
+
+        from persistent._compat import IterableUserDict
+        state = IterableUserDict({'foobar': [1, 2]})
+
+        inst1.__setstate__(state)
+        self.assertTrue(hasattr(inst1, 'foobar'))
+
     def test___reduce__(self):
-        from persistent._compat import copy_reg
         inst = self._makeOne()
         first, second, third = inst.__reduce__()
         self.assertTrue(first is copy_reg.__newobj__)
@@ -1004,7 +1050,6 @@ class _Persistent_Base(object):
         self.assertEqual(third, None)
 
     def test___reduce__w_subclass_having_getnewargs(self):
-        from persistent._compat import copy_reg
         class Derived(self._getTargetClass()):
             def __getnewargs__(self):
                 return ('a', 'b')
@@ -1015,7 +1060,6 @@ class _Persistent_Base(object):
         self.assertEqual(third, {})
 
     def test___reduce__w_subclass_having_getstate(self):
-        from persistent._compat import copy_reg
         class Derived(self._getTargetClass()):
             def __getstate__(self):
                 return {}
@@ -1026,7 +1070,6 @@ class _Persistent_Base(object):
         self.assertEqual(third, {})
 
     def test___reduce__w_subclass_having_getnewargs_and_getstate(self):
-        from persistent._compat import copy_reg
         class Derived(self._getTargetClass()):
             def __getnewargs__(self):
                 return ('a', 'b')
@@ -1038,10 +1081,21 @@ class _Persistent_Base(object):
         self.assertEqual(second, (Derived, 'a', 'b'))
         self.assertEqual(third, {'foo': 'bar'})
 
+    def _get_cucumber(self, name):
+        # Checks that it's actually a subclass of what we're testing;
+        # if it isn't, the test is skipped. The cucumbers module can
+        # only subclass either the C or Python implementation, not
+        # both
+        from persistent.tests import cucumbers
+        cls = getattr(cucumbers, name)
+        if not issubclass(cls, self._getTargetClass()):
+            self.skipTest("Cucumber is not correct implementation")
+        return cls
+
     def test_pickle_roundtrip_simple(self):
         import pickle
         # XXX s.b. 'examples'
-        from persistent.tests.cucumbers import Simple
+        Simple = self._get_cucumber('Simple')
         inst = Simple('testing')
         copy = pickle.loads(pickle.dumps(inst))
         self.assertEqual(copy, inst)
@@ -1052,7 +1106,7 @@ class _Persistent_Base(object):
     def test_pickle_roundtrip_w_getnewargs_and_getstate(self):
         import pickle
         # XXX s.b. 'examples'
-        from persistent.tests.cucumbers import Custom
+        Custom = self._get_cucumber('Custom')
         inst = Custom('x', 'y')
         copy = pickle.loads(pickle.dumps(inst))
         self.assertEqual(copy, inst)
@@ -1063,7 +1117,7 @@ class _Persistent_Base(object):
     def test_pickle_roundtrip_w_slots_missing_slot(self):
         import pickle
         # XXX s.b. 'examples'
-        from persistent.tests.cucumbers import SubSlotted
+        SubSlotted = self._get_cucumber('SubSlotted')
         inst = SubSlotted('x', 'y', 'z')
         copy = pickle.loads(pickle.dumps(inst))
         self.assertEqual(copy, inst)
@@ -1074,7 +1128,7 @@ class _Persistent_Base(object):
     def test_pickle_roundtrip_w_slots_filled_slot(self):
         import pickle
         # XXX s.b. 'examples'
-        from persistent.tests.cucumbers import SubSlotted
+        SubSlotted = self._get_cucumber('SubSlotted')
         inst = SubSlotted('x', 'y', 'z')
         inst.s4 = 'a'
         copy = pickle.loads(pickle.dumps(inst))
@@ -1086,7 +1140,7 @@ class _Persistent_Base(object):
     def test_pickle_roundtrip_w_slots_and_empty_dict(self):
         import pickle
         # XXX s.b. 'examples'
-        from persistent.tests.cucumbers import SubSubSlotted
+        SubSubSlotted = self._get_cucumber('SubSubSlotted')
         inst = SubSubSlotted('x', 'y', 'z')
         copy = pickle.loads(pickle.dumps(inst))
         self.assertEqual(copy, inst)
@@ -1097,7 +1151,7 @@ class _Persistent_Base(object):
     def test_pickle_roundtrip_w_slots_and_filled_dict(self):
         import pickle
         # XXX s.b. 'examples'
-        from persistent.tests.cucumbers import SubSubSlotted
+        SubSubSlotted = self._get_cucumber('SubSubSlotted')
         inst = SubSubSlotted('x', 'y', 'z', foo='bar', baz='bam')
         inst.s4 = 'a'
         copy = pickle.loads(pickle.dumps(inst))
@@ -1321,22 +1375,39 @@ class _Persistent_Base(object):
 
     def test__p_invalidate_from_changed_w_slots(self):
         class Derived(self._getTargetClass()):
-            __slots__ = ('myattr1', 'myattr2', 'unset')
+            __slots__ = {
+                'myattr1': 'value1',
+                'myattr2': 'value2',
+                'unset': None
+            }
+
             def __init__(self):
                 self.myattr1 = 'value1'
                 self.myattr2 = 'value2'
+
         inst, jar, OID = self._makeOneWithJar(Derived)
         inst._p_activate()
         inst._p_changed = True
         jar._loaded = []
         jar._registered = []
-        self.assertEqual(Derived.myattr1.__get__(inst), 'value1')
-        self.assertEqual(Derived.myattr2.__get__(inst), 'value2')
+        for slot, expected_value in Derived.__slots__.items():
+            slot_descriptor = getattr(Derived, slot)
+            if expected_value:
+                self.assertEqual(slot_descriptor.__get__(inst), expected_value)
+            else:
+                with self.assertRaises(AttributeError):
+                    slot_descriptor.__get__(inst)
+
         inst._p_invalidate()
         self.assertEqual(inst._p_status, 'ghost')
         self.assertEqual(list(jar._loaded), [])
-        self.assertRaises(AttributeError, lambda: Derived.myattr1.__get__(inst))
-        self.assertRaises(AttributeError, lambda: Derived.myattr2.__get__(inst))
+
+        for slot in Derived.__slots__:
+            __traceback_info__ = slot, inst
+            slot_descriptor = getattr(Derived, slot)
+            with self.assertRaises(AttributeError):
+                slot_descriptor.__get__(inst)
+
         self.assertEqual(list(jar._loaded), [])
         self.assertEqual(list(jar._registered), [])
 
@@ -1370,6 +1441,26 @@ class _Persistent_Base(object):
         self.assertEqual(list(jar._loaded), [])
         self.assertEqual(list(jar._registered), [])
 
+    def test_p_invalidate_with_slots_broken_jar(self):
+        # If jar.setstate() raises a POSKeyError (or any error)
+        # clearing an object with unset slots doesn't result in a
+        # SystemError, the original error is propagated
+
+        class Derived(self._getTargetClass()):
+            __slots__ = ('slot1',)
+
+        # Pre-cache in __slotnames__; cpersistent goes directly for this
+        # and avoids a call to copy_reg. (If it calls the python code in
+        # copy_reg, the pending exception will be immediately propagated by
+        # copy_reg, not by us.)
+        copy_reg._slotnames(Derived)
+
+        inst, jar, OID = self._makeOneWithJar(Derived, broken_jar=True)
+        inst._p_invalidate()
+        self.assertEqual(inst._p_status, 'ghost')
+        self.assertRaises(NotImplementedError, inst._p_activate)
+
+
     def test__p_invalidate_from_sticky(self):
         inst, jar, OID = self._makeOneWithJar()
         inst._p_activate() # XXX
@@ -1433,8 +1524,7 @@ class _Persistent_Base(object):
         self._checkMRU(jar, [OID])
 
     def test__p_setattr_w__p__name(self):
-        from persistent.timestamp import _makeOctets
-        SERIAL = _makeOctets('\x01' * 8)
+        SERIAL = b'\x01' * 8
         inst, jar, OID = self._makeOneWithJar()
         inst._p_deactivate()
         self.assertTrue(inst._p_setattr('_p_serial', SERIAL))
@@ -1483,14 +1573,12 @@ class _Persistent_Base(object):
         # object stays in the up-to-date state.
         # It shouldn't change to the modified state, because it won't
         # be saved when the transaction commits.
-        from persistent._compat import _b
         class P(self._getTargetClass()):
             def __init__(self):
                 self.x = 0
-            def inc(self):
-                self.x += 1
+
         p = P()
-        p._p_oid = _b('1')
+        p._p_oid = b'1'
         p._p_jar = self._makeBrokenJar()
         self.assertEqual(p._p_state, 0)
         self.assertEqual(p._p_jar.called, 0)
@@ -1503,14 +1591,11 @@ class _Persistent_Base(object):
     def test__p_activate_w_broken_jar(self):
         # Make sure that exceptions that occur inside the data manager's
         # ``setstate()`` method propagate out to the caller.
-        from persistent._compat import _b
         class P(self._getTargetClass()):
             def __init__(self):
                 self.x = 0
-            def inc(self):
-                self.x += 1
         p = P()
-        p._p_oid = _b('1')
+        p._p_oid = b'1'
         p._p_jar = self._makeBrokenJar()
         p._p_deactivate()
         self.assertEqual(p._p_state, -1)
@@ -1564,23 +1649,21 @@ class _Persistent_Base(object):
         class subclass(self._getTargetClass()):
             _v_setattr_called = False
             def __setattr__(self, name, value):
-                object.__setattr__(self, '_v_setattr_called', True)
-                super(subclass,self).__setattr__(name, value)
+                raise AssertionError("Should not be called")
         inst = subclass()
-        self.assertEqual(object.__getattribute__(inst,'_v_setattr_called'), False)
+        self.assertEqual(object.__getattribute__(inst, '_v_setattr_called'), False)
 
     def test_can_set__p_attrs_if_subclass_denies_setattr(self):
-        from persistent._compat import _b
         # ZODB defines a PersistentBroken subclass that only lets us
         # set things that start with _p, so make sure we can do that
         class Broken(self._getTargetClass()):
             def __setattr__(self, name, value):
                 if name.startswith('_p_'):
-                    super(Broken,self).__setattr__(name, value)
+                    super(Broken, self).__setattr__(name, value)
                 else:
-                    raise TypeError("Can't change broken objects")
+                    raise AssertionError("Can't change broken objects")
 
-        KEY = _b('123')
+        KEY = b'123'
         jar = self._makeJar()
 
         broken = Broken()
@@ -1599,11 +1682,269 @@ class _Persistent_Base(object):
         p._p_invalidate()
         self.assertTrue(p.deactivated)
 
+
+    def test_new_ghost_success_not_already_ghost_dict(self):
+        # https://github.com/zopefoundation/persistent/issues/49
+        # calling new_ghost on an object that already has state just changes
+        # its flags, it doesn't destroy the state.
+        from persistent.interfaces import GHOST
+        from persistent.interfaces import UPTODATE
+        class TestPersistent(self._getTargetClass()):
+            pass
+        KEY = b'123'
+        jar = self._makeJar()
+        cache = self._makeRealCache(jar)
+        candidate = TestPersistent()
+
+        candidate.set_by_new = 1
+        self.assertEqual(candidate._p_state, UPTODATE)
+        cache.new_ghost(KEY, candidate)
+
+        self.assertIs(cache.get(KEY), candidate)
+        self.assertEqual(candidate._p_oid, KEY)
+        self.assertEqual(candidate._p_state, GHOST)
+        self.assertEqual(candidate.set_by_new, 1)
+
+    def test_new_ghost_success_not_already_ghost_slot(self):
+        # https://github.com/zopefoundation/persistent/issues/49
+        # calling new_ghost on an object that already has state just changes
+        # its flags, it doesn't destroy the state.
+        from persistent.interfaces import GHOST
+        from persistent.interfaces import UPTODATE
+        class TestPersistent(self._getTargetClass()):
+            __slots__ = ('set_by_new', '__weakref__')
+        KEY = b'123'
+        jar = self._makeJar()
+        cache = self._makeRealCache(jar)
+        candidate = TestPersistent()
+        candidate.set_by_new = 1
+        self.assertEqual(candidate._p_state, UPTODATE)
+        cache.new_ghost(KEY, candidate)
+
+        self.assertIs(cache.get(KEY), candidate)
+        self.assertEqual(candidate._p_oid, KEY)
+        self.assertEqual(candidate._p_state, GHOST)
+        self.assertEqual(candidate.set_by_new, 1)
+
+    # The number 12345678 as a p64, 8-byte string
+    _PACKED_OID = b'\x00\x00\x00\x00\x00\xbcaN'
+    # The number 12345678 printed in hex
+    _HEX_OID = '0xbc614e'
+
+    def _normalize_repr(self, r):
+        # addresses
+        r = re.sub(r'at 0x[0-9a-fA-F]*', 'at 0xdeadbeef', r)
+        # Python 3.7 removed the trailing , in exception reprs
+        r = r.replace("',)", "')")
+        return r
+
+    def _normalized_repr(self, o):
+        return self._normalize_repr(repr(o))
+
+    def test_repr_no_oid_no_jar(self):
+        p = self._makeOne()
+        result = self._normalized_repr(p)
+        self.assertEqual(result, '<persistent.Persistent object at 0xdeadbeef>')
+
+    def test_repr_no_oid_in_jar(self):
+        p = self._makeOne()
+
+        class Jar(object):
+            def __repr__(self):
+                return '<SomeJar>'
+
+        p._p_jar = Jar()
+
+        result = self._normalized_repr(p)
+        self.assertEqual(
+            result,
+            "<persistent.Persistent object at 0xdeadbeef in <SomeJar>>")
+
+    def test_repr_oid_no_jar(self):
+        p = self._makeOne()
+        p._p_oid = self._PACKED_OID
+
+        result = self._normalized_repr(p)
+        self.assertEqual(
+            result,
+            "<persistent.Persistent object at 0xdeadbeef oid " + self._HEX_OID + ">")
+
+    def test_64bit_oid(self):
+        import struct
+        p = self._makeOne()
+        oid_value = 2 << 62
+        self.assertEqual(oid_value.bit_length(), 64)
+        oid = struct.pack(">Q", oid_value)
+        self.assertEqual(oid, b'\x80\x00\x00\x00\x00\x00\x00\x00')
+
+        p._p_oid = oid
+        result = self._normalized_repr(p)
+        self.assertEqual(
+            result,
+            '<persistent.Persistent object at 0xdeadbeef oid 0x8000000000000000>'
+        )
+
+    def test_repr_no_oid_repr_jar_raises_exception(self):
+        p = self._makeOne()
+
+        class Jar(object):
+            def __repr__(self):
+                raise Exception('jar repr failed')
+
+        p._p_jar = Jar()
+
+        result = self._normalized_repr(p)
+        self.assertEqual(
+            result,
+            "<persistent.Persistent object at 0xdeadbeef in Exception('jar repr failed')>")
+
+
+    def test_repr_oid_raises_exception_no_jar(self):
+        p = self._makeOne()
+
+        class BadOID(bytes):
+            def __repr__(self):
+                raise Exception("oid repr failed")
+
+        # Our OID is bytes, 8 bytes long. We don't call its repr.
+        p._p_oid = BadOID(self._PACKED_OID)
+
+        result = self._normalized_repr(p)
+        self.assertEqual(
+            result,
+            "<persistent.Persistent object at 0xdeadbeef oid " + self._HEX_OID + ">")
+
+        # Anything other than 8 bytes, though, we do.
+        p._p_oid = BadOID(b'1234567')
+
+        result = self._normalized_repr(p)
+        self.assertEqual(
+            result,
+            "<persistent.Persistent object at 0xdeadbeef oid Exception('oid repr failed')>")
+
+
+    def test_repr_oid_and_jar_raise_exception(self):
+        p = self._makeOne()
+
+        class BadOID(bytes):
+            def __repr__(self):
+                raise Exception("oid repr failed")
+        p._p_oid = BadOID(b'1234567')
+
+        class Jar(object):
+            def __repr__(self):
+                raise Exception('jar repr failed')
+
+        p._p_jar = Jar()
+
+
+        result = self._normalized_repr(p)
+        self.assertEqual(
+            result,
+            "<persistent.Persistent object at 0xdeadbeef oid Exception('oid repr failed')"
+            " in Exception('jar repr failed')>")
+
+    def test_repr_no_oid_repr_jar_raises_baseexception(self):
+        p = self._makeOne()
+
+        class Jar(object):
+            def __repr__(self):
+                raise BaseException('jar repr failed')
+
+        p._p_jar = Jar()
+        with self.assertRaisesRegex(BaseException, 'jar repr failed'):
+            repr(p)
+
+    def test_repr_oid_raises_baseexception_no_jar(self):
+        p = self._makeOne()
+
+        class BadOID(bytes):
+            def __repr__(self):
+                raise BaseException("oid repr failed")
+        p._p_oid = BadOID(b'12345678')
+
+        # An 8 byte byte string doesn't have repr called.
+        repr(p)
+
+        # Anything other does.
+        p._p_oid = BadOID(b'1234567')
+        with self.assertRaisesRegex(BaseException, 'oid repr failed'):
+            repr(p)
+
+    def test_repr_oid_and_jar(self):
+        p = self._makeOne()
+        p._p_oid = self._PACKED_OID
+
+        class Jar(object):
+            def __repr__(self):
+                return '<SomeJar>'
+
+        p._p_jar = Jar()
+
+        result = self._normalized_repr(p)
+        self.assertEqual(
+            result,
+            "<persistent.Persistent object at 0xdeadbeef oid " + self._HEX_OID + " in <SomeJar>>")
+
+    def test__p_repr(self):
+        class P(self._getTargetClass()):
+            def _p_repr(self):
+                return "Override"
+        p = P()
+        self.assertEqual("Override", repr(p))
+
+    def test__p_repr_exception(self):
+        class P(self._getTargetClass()):
+            def _p_repr(self):
+                raise Exception("_p_repr failed")
+        p = P()
+        result = self._normalized_repr(p)
+        self.assertEqual(
+            result,
+            "<persistent.tests.test_persistence.P object at 0xdeadbeef"
+            " _p_repr Exception('_p_repr failed')>")
+
+        p._p_oid = self._PACKED_OID
+        result = self._normalized_repr(p)
+        self.assertEqual(
+            result,
+            "<persistent.tests.test_persistence.P object at 0xdeadbeef oid " + self._HEX_OID
+            + " _p_repr Exception('_p_repr failed')>")
+
+        class Jar(object):
+            def __repr__(self):
+                return '<SomeJar>'
+
+        p._p_jar = Jar()
+        result = self._normalized_repr(p)
+        self.assertEqual(
+            result,
+            "<persistent.tests.test_persistence.P object at 0xdeadbeef oid " + self._HEX_OID
+            + " in <SomeJar> _p_repr Exception('_p_repr failed')>")
+
+    def test__p_repr_in_instance_ignored(self):
+        class P(self._getTargetClass()):
+            pass
+        p = P()
+        p._p_repr = lambda: "Instance"
+        result = self._normalized_repr(p)
+        self.assertEqual(result,
+                         '<persistent.tests.test_persistence.P object at 0xdeadbeef>')
+
+    def test__p_repr_baseexception(self):
+        class P(self._getTargetClass()):
+            def _p_repr(self):
+                raise BaseException("_p_repr failed")
+        p = P()
+        with self.assertRaisesRegex(BaseException, '_p_repr failed'):
+            repr(p)
+
 class PyPersistentTests(unittest.TestCase, _Persistent_Base):
 
     def _getTargetClass(self):
-        from persistent.persistence import Persistent
-        return Persistent
+        from persistent.persistence import PersistentPy
+        assert PersistentPy.__module__ == 'persistent.persistence', PersistentPy.__module__
+        return PersistentPy
 
     def _makeCache(self, jar):
 
@@ -1627,6 +1968,10 @@ class PyPersistentTests(unittest.TestCase, _Persistent_Base):
 
         return _Cache(jar)
 
+    def _makeRealCache(self, jar):
+        from persistent.picklecache import PickleCachePy as PickleCache
+        return PickleCache(jar, 10)
+
     def _checkMRU(self, jar, value):
         self.assertEqual(list(jar._cache._mru), value)
 
@@ -1639,26 +1984,24 @@ class PyPersistentTests(unittest.TestCase, _Persistent_Base):
         # but because they are newly created, they aren't in the
         # pickle cache yet.
         # Nothing should blow up when this happens
-        from persistent._compat import _b
-        KEY = _b('123')
+        KEY = b'123'
         jar = self._makeJar()
         c1 = self._makeOne()
         c1._p_oid = KEY
         c1._p_jar = jar
-        orig_mru = jar._cache.mru
+
         def mru(oid):
             # Mimic what the real cache does
             if oid not in jar._cache._mru:
                 raise KeyError(oid)
-            orig_mru(oid)
+            raise AssertionError("Should never get here")
         jar._cache.mru = mru
         c1._p_accessed()
         self._checkMRU(jar, [])
 
     def test_accessed_invalidated_with_jar_and_oid_but_no_cache(self):
         # This scenario arises in ZODB tests where the jar is faked
-        from persistent._compat import _b
-        KEY = _b('123')
+        KEY = b'123'
         class Jar(object):
             accessed = False
             def __getattr__(self, name):
@@ -1707,52 +2050,41 @@ class PyPersistentTests(unittest.TestCase, _Persistent_Base):
         inst._Persistent__flags = None
         inst._p_accessed()
 
-_add_to_suite = [PyPersistentTests]
 
-if not os.environ.get('PURE_PYTHON'):
-    try:
-        from persistent import cPersistence
-    except ImportError:
-        pass
-    else:
-        class CPersistentTests(unittest.TestCase, _Persistent_Base):
+@skipIfNoCExtension
+class CPersistentTests(unittest.TestCase, _Persistent_Base):
 
-            def _getTargetClass(self):
-                from persistent.cPersistence import Persistent
-                return Persistent
+    def _getTargetClass(self):
+        from persistent._compat import _c_optimizations_available as get_c
+        return get_c()['persistent.persistence'].Persistent
 
-            def _checkMRU(self, jar, value):
-                pass # Figure this out later
+    def _checkMRU(self, jar, value):
+        pass # Figure this out later
 
-            def _clearMRU(self, jar):
-                pass # Figure this out later
+    def _clearMRU(self, jar):
+        pass # Figure this out later
 
-            def _makeCache(self, jar):
-                from persistent.cPickleCache import PickleCache
-                return PickleCache(jar)
+    def _makeCache(self, jar):
+        from persistent._compat import _c_optimizations_available as get_c
+        PickleCache = get_c()['persistent.picklecache'].PickleCache
+        return PickleCache(jar)
 
-        _add_to_suite.append(CPersistentTests)
 
-        class Test_simple_new(unittest.TestCase):
+@skipIfNoCExtension
+class Test_simple_new(unittest.TestCase):
 
-            def _callFUT(self, x):
-                from persistent.cPersistence import simple_new
-                return simple_new(x)
+    def _callFUT(self, x):
+        from persistent._compat import _c_optimizations_available as get_c
+        simple_new = get_c()['persistent.persistence'].simple_new
+        return simple_new(x)
 
-            def test_w_non_type(self):
-                self.assertRaises(TypeError, self._callFUT, '')
+    def test_w_non_type(self):
+        self.assertRaises(TypeError, self._callFUT, '')
 
-            def test_w_type(self):
-                import sys
-                TO_CREATE = [type, list, tuple, object]
-                # Python 3.3 segfaults when destroying a dict created via
-                # PyType_GenericNew.  See http://bugs.python.org/issue16676
-                if sys.version_info < (3, 3):
-                    TO_CREATE.append(dict)
-                for typ in TO_CREATE:
-                    self.assertTrue(isinstance(self._callFUT(typ), typ))
-
-        _add_to_suite.append(Test_simple_new)
+    def test_w_type(self):
+        TO_CREATE = [type, list, tuple, object, dict]
+        for typ in TO_CREATE:
+            self.assertTrue(isinstance(self._callFUT(typ), typ))
 
 def test_suite():
-    return unittest.TestSuite([unittest.makeSuite(x) for x in _add_to_suite])
+    return unittest.defaultTestLoader.loadTestsFromName(__name__)
